@@ -23,6 +23,74 @@ class DataInterchange():
         self.comments=[]
         pass
     
+    @staticmethod 
+    def ofDict(pDict,context="context"):
+        dif=DataInterchange()
+        dif.addTriple(Triple(context,"isA","Context"))
+        dif.addTriple(Triple(context,"name","it"))
+        dif.addSchemaFromDict(pDict,context,context,"context")
+        return dif
+    
+    def addLink(self,name,source,target,sourceRole,targetRole,sourceMultiple,targetMultiple):
+        self.addTriple(Triple(name,"isA","TopicLink"))
+        self.addTriple(Triple(name,"name","it"))
+        self.addTriple(Triple(target,"target","it"))
+        self.addTriple(Triple(source,"source","it"))
+        self.addTriple(Triple(targetRole,"targetRole","it"))
+        self.addTriple(Triple(sourceRole,"sourceRole","it"))
+        self.addTriple(Triple(sourceMultiple,"sourceMultiple","it"))
+        self.addTriple(Triple(targetMultiple,"targetMultiple","it"))
+        
+    def addTopic(self,name,context):
+        topicId=self.fixId(name)
+        self.addTriple(Triple(topicId,"isA","Topic"))
+        self.addTriple(Triple(name,"name","it"))
+        self.addTriple(Triple(context,"context","it"))
+        
+    def fixId(self,name):
+        fixed=re.sub(r"[#@]","",name)
+        return fixed
+        
+        
+    def addSchemaFromDict(self,pDict,context,parent,parentName):    
+        '''
+        add schema information from the given dict
+        '''
+        if not isinstance(pDict,dict):
+            return
+        # make sure we work on properties first then other topics
+        sortedKeys=sorted(pDict.keys(), key=lambda x: isinstance(pDict[x],dict))
+        # loop over nodes
+        for key in sortedKeys:
+            value=pDict[key]
+            # is the subnode a Topic or a Property?
+            if isinstance(value,dict):
+                # if there is an intermediate node
+                # then there is a topic link
+                # eg. workshop - events - event
+                # https://stackoverflow.com/questions/21062781/shortest-way-to-get-first-item-of-ordereddict-in-python-3
+                if len(value)==1 and isinstance(list(value.values())[0],list):
+                    listNode=list(value.values())[0]
+                    firstListNode=listNode[0]
+                    listKey=list(value.keys())[0]
+                    self.addTopic(listKey,context)
+                    self.addSchemaFromDict(firstListNode,context,listKey, "Topic")
+                    self.addLink(key, parent, listKey, "", key, False, len(listNode)>1)
+                else:
+                    # standalone topic
+                    linkKey="%s%s" % (parent,key)
+                    self.addTopic(key,context)
+                    self.addSchemaFromDict(value,context,key,"Topic")
+                    if parentName!="context":
+                        self.addLink(linkKey,parent,key,"","",False,False)
+            else:
+                propId=self.fixId(key)
+                self.addTriple(Triple(propId,"isA","Property"))
+                self.addTriple(Triple(key,"name","it"))  
+                valueType=type(value).__name__
+                self.addTriple(Triple(valueType,"type","it"))  
+                self.addTriple(Triple(parent,parentName,"it"))
+    
     def addTriple(self,triple):
         '''
         add the given triple
@@ -38,32 +106,53 @@ class DataInterchange():
         '''
         self.comments.append(comment)
         
+    def asSiDIF(self):
+        '''
+        convert me to SiDIF notation
+        '''
+        sidifStr=""
+        for triple in self.triples:
+            sidif=triple.asSiDIF()
+            sidifStr+=f"{sidif}\n"
+        return sidifStr
+    
     def toDictOfDicts(self):
         '''
         convert me to a dict of dicts 
         following the "it" semantics
         
         e.g.
-            JohnDoe isA Person
-            "John" is firstName of it
-            "Doe"  is lastName of it
-            35 is age of it
+        
+        .. code-block:: python
+        
+               JohnDoe isA Person
+               "John" is firstName of it
+               "Doe"  is lastName of it
+               35 is age of it
             
         will have a pseudo - triple representation of
+        
+        .. code-block:: python
+        
             JohnDoe isA Person
             John firstName it
             Doe lastName it
             35 age it
             
-        leading to a dict {
-          'JohnDoe': { 
-              'isA': Person, 
-              'firstName': John, 
-              'lastName': 'Doe'
-              'age': 35
-        }
+        leading to a dict 
         
-        Return:
+        .. code-block:: python
+        
+            {
+               'JohnDoe': { 
+                  'isA': Person, 
+                  'firstName': John, 
+                   'lastName': 'Doe'
+                  'age': 35
+                }
+            }    
+        
+        Returns:
             dict: the dict of dicts representation of the triples found
         '''
         # the dict of dicts
@@ -128,6 +217,22 @@ class Triple():
         d="%s(%s)" % (value,type(value).__name__)
         return d
     
+    def asLiteral(self,value):
+        if isinstance(value,str):
+            return '"%s"' % value
+        elif isinstance(value,bool):
+            return "true" if value else "false"
+        else:
+            return "%s" % value
+    
+    def asSiDIF(self):
+        if self.o=="it":
+            literal=self.asLiteral(self.s)
+            line='%s is %s of it' % (literal,self.p)
+        else:
+            line="%s %s %s" % (self.s,self.p,self.o)
+        return line
+    
     def __str__(self):
         text="{%s,%s,%s}" % (self.dump(self.s),self.dump(self.p),self.dump(self.o))
         return text
@@ -197,7 +302,7 @@ class SiDIFParser(object):
         r"(?::\d{2,5})?"
         # resource path (optional)
         r"(?:[/?#]\S*)?")
-        compiled=re.compile(uriRegexp,re.RegexFlag.I |re.RegexFlag.UNICODE)
+        compiled=re.compile(uriRegexp,re.IGNORECASE |re.UNICODE)
         return compiled
     
     def convertToTime(self,tokenStr,location,token):
@@ -213,9 +318,9 @@ class SiDIFParser(object):
         except ValueError as ve:
             raise ParseFatalException(tokenStr, location, str(ve))
         
-    def convertToBoolean(self,tokenStr,location,token):
+    def convertToBoolean(self,tokenStr:str,location,token):
         '''
-        convert the token to a boolean
+        convert the given token to a boolean
         '''
         try:
             tokenStr=token[0]
@@ -228,7 +333,7 @@ class SiDIFParser(object):
         # https://stackoverflow.com/questions/13393432/raise-a-custom-exception-in-pyparsing
         raise ParseFatalException(tokenStr, location, "invalid boolean %s:%s" % (tokenStr,msg))
     
-    def handleDateTimeLiteral(self,tokenStr,location,group):
+    def handleDateTimeLiteral(self,tokenStr:str,location,group):
         '''
         handle a date time literal
         '''
@@ -278,9 +383,9 @@ class SiDIFParser(object):
         return comment
         
     def handleGroup(self,_tokenStr,_location,tokens):
-        tokenName=tokens.getName()
+        _tokenName=tokens.getName()
         token=tokens[0]
-        innerName=token.getName()
+        _innerName=token.getName()
         inner=token[0]
         return inner
     
